from .field_utils import *  # noqa: 401
from .model_generator.generator import BaseClassEnum, ModelGenerator  # noqa: 401
from .schema_generator import AvroModel  # noqa: 401
from .types import *  # noqa: 401



import dataclasses
import datetime
import sys
import typing

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated  # type: ignore # pragma: no cover

if sys.version_info >= (3, 10):
    from types import UnionType  # pragma: no cover
else:
    UnionType = None  # pragma: no cover

T = typing.TypeVar("T")
JsonDict = typing.Dict[str, typing.Any]

CUSTOM_TYPES = (
    "Fixed",
    "Decimal",
    "Int32",
    "Float32",
    "TimeMicro",
    "DateTimeMicro",
)

__all__ = CUSTOM_TYPES


class MissingSentinel(typing.Generic[T]):
    """
    Class to detect when a field is not initialized
    """

    ...


@dataclasses.dataclass(frozen=True)
class Fixed(typing.Generic[T]):
    """
    Represents an Avro Fixed type

    size (int): Specifying the number of bytes per value
    """

    size: int
    default: typing.Any = dataclasses.field(default=MissingSentinel)
    namespace: typing.Optional[str] = None
    aliases: typing.Optional[typing.List] = None
    _dataclasses_custom_type: str = "Fixed"

    def __repr__(self) -> str:
        return f"{self.size}"


class Decimal:
    """
    Represents an Avro Decimal type
    precision (int): Specifying the number precision
    scale(int): Specifying the number scale. Default 0
    """

    def __init__(
        self,
        *,
        precision: int,
        scale: int = 0,
        default: typing.Any = MissingSentinel,
        aliases: typing.Optional[typing.List] = None,
    ) -> None:
        self.precision = precision
        self.scale = scale
        self.default = default
        self.aliases = aliases

    # Decimal serializes to bytes, which doesn't support namespace

    def __repr__(self) -> str:
        return f"Decimal('{self.default}')"


Int32 = Annotated[int, "Int32"]
Float32 = Annotated[float, "Float32"]
TimeMicro = Annotated[datetime.time, "TimeMicro"]
DateTimeMicro = Annotated[datetime.datetime, "DateTimeMicro"]



import datetime
import decimal
import enum
import typing
import uuid

from . import types

__all__ = [
    "BOOLEAN",
    "NULL",
    "INT",
    "FLOAT",
    "LONG",
    "DOUBLE",
    "BYTES",
    "STRING",
    "ARRAY",
    "ENUM",
    "MAP",
    "FIXED",
    "DATE",
    "UUID",
    "DECIMAL",
    "RECORD",
    "TIME_MILLIS",
    "TIME_MICROS",
    "TIMESTAMP_MILLIS",
    "TIMESTAMP_MICROS",
    "LOGICAL_DATE",
    "LOGICAL_TIME_MILIS",
    "LOGICAL_TIME_MICROS",
    "LOGICAL_DATETIME_MILIS",
    "LOGICAL_DATETIME_MICROS",
    "LOGICAL_UUID",
    "PYTHON_TYPE_TO_AVRO",
    "PYTHON_INMUTABLE_TYPES",
    "PYTHON_PRIMITIVE_CONTAINERS",
    "PYTHON_LOGICAL_TYPES",
    "PYTHON_PRIMITIVE_TYPES",
    "PRIMITIVE_AND_LOGICAL_TYPES",
    "PythonImmutableTypes",
]

TIME_MILLIS = "time-millis"
TIME_MICROS = "time-micros"
TIMESTAMP_MILLIS = "timestamp-millis"
TIMESTAMP_MICROS = "timestamp-micros"

BOOLEAN = "boolean"
NULL = "null"
INT = "int"
FLOAT = "float"
LONG = "long"
DOUBLE = "double"
BYTES = "bytes"
STRING = "string"
ARRAY = "array"
ENUM = "enum"
MAP = "map"
FIXED = "fixed"
DATE = "date"
UUID = "uuid"
DECIMAL = "decimal"
RECORD = "record"
LOGICAL_DATE = {"type": INT, "logicalType": DATE}
LOGICAL_TIME_MILIS = {"type": INT, "logicalType": TIME_MILLIS}
LOGICAL_TIME_MICROS = {"type": LONG, "logicalType": TIME_MICROS}
LOGICAL_DATETIME_MILIS = {"type": LONG, "logicalType": TIMESTAMP_MILLIS}
LOGICAL_DATETIME_MICROS = {"type": LONG, "logicalType": TIMESTAMP_MICROS}
LOGICAL_UUID = {"type": STRING, "logicalType": UUID}

PYTHON_TYPE_TO_AVRO = {
    bool: BOOLEAN,
    type(None): NULL,
    int: LONG,
    float: DOUBLE,
    bytes: BYTES,
    str: STRING,
    list: {"type": ARRAY},
    tuple: {"type": ARRAY},
    dict: {"type": MAP},
    enum.Enum: {"type": ENUM},
    types.Fixed: {"type": FIXED},
    types.Int32: INT,
    types.Float32: FLOAT,
    datetime.date: {"type": INT, "logicalType": DATE},
    datetime.time: {"type": INT, "logicalType": TIME_MILLIS},
    datetime.datetime: {"type": LONG, "logicalType": TIMESTAMP_MILLIS},
    uuid.uuid4: {"type": STRING, "logicalType": UUID},
}

# excluding tuple because is a container
PYTHON_INMUTABLE_TYPES = (str, int, types.Int32, types.Float32, bool, float, bytes, type(None))
PYTHON_PRIMITIVE_CONTAINERS = (list, tuple, dict)

PYTHON_LOGICAL_TYPES = (
    datetime.date,
    datetime.time,
    types.TimeMicro,
    datetime.datetime,
    types.DateTimeMicro,
    uuid.uuid4,
    uuid.UUID,
    decimal.Decimal,
)

PYTHON_PRIMITIVE_TYPES = PYTHON_INMUTABLE_TYPES + PYTHON_PRIMITIVE_CONTAINERS
PRIMITIVE_AND_LOGICAL_TYPES = PYTHON_INMUTABLE_TYPES + PYTHON_LOGICAL_TYPES

PythonImmutableTypes = typing.Union[
    str,
    int,
    types.Int32,
    bool,
    float,
    types.Float32,
    list,
    tuple,
    dict,
    datetime.date,
    datetime.time,
    datetime.datetime,
    uuid.UUID,
    decimal.Decimal,
]



import dataclasses
import typing
from datetime import datetime

from pytz import utc

from .types import JsonDict

try:
    import faust
except ImportError:  # pragma: no cover
    faust = None  # type: ignore # pragma: no cover

try:
    import pydantic  # pragma: no cover
except ImportError:  # type: ignore # pragma: no cover
    pydantic = None  # type: ignore # pragma: no cover


def is_pydantic_model(klass: type) -> bool:
    if pydantic is not None:
        return issubclass(klass, pydantic.BaseModel)
    return False


def is_dataclass_or_pydantic_model(type: type) -> bool:
    return dataclasses.is_dataclass(type) or is_pydantic_model(type)


def is_faust_model(klass: type) -> bool:
    if faust is not None:
        return issubclass(klass, faust.Record)
    return False


def is_union(a_type: type) -> bool:
    """
    Given a python type, return True if is typing.Union, otherwise False

    Arguments:
        a_type (typing.Any): python type

    Returns:
        bool
    """
    return isinstance(a_type, typing._GenericAlias) and a_type.__origin__ is typing.Union  # type: ignore


def is_self_referenced(a_type: type) -> bool:
    """
    Given a python type, return True if is self referenced, meaning
    that is instance of typing.ForwardRef, otherwise False

    Arguments:
        a_type (typing.Any): python type

    Returns:
        bool

    Example:
        a_type = typing.Type["User"]]

        is_self_referenced(a_type) # True
    """
    return (
        isinstance(a_type, typing._GenericAlias)  # type: ignore
        and a_type.__args__
        and isinstance(a_type.__args__[0], typing.ForwardRef)  # type: ignore
    )


@dataclasses.dataclass
class SchemaMetadata:
    schema_name: typing.Optional[str] = None
    schema_doc: typing.Union[bool, str] = True
    namespace: typing.Optional[typing.List[str]] = None
    aliases: typing.Optional[typing.List[str]] = None
    alias_nested_items: typing.Dict[str, str] = dataclasses.field(default_factory=dict)
    dacite_config: typing.Optional[JsonDict] = None

    @classmethod
    def create(cls: typing.Type["SchemaMetadata"], klass: type) -> typing.Any:
        return cls(
            schema_name=getattr(klass, "schema_name", None),
            schema_doc=getattr(klass, "schema_doc", True),
            namespace=getattr(klass, "namespace", None),
            aliases=getattr(klass, "aliases", None),
            alias_nested_items=getattr(klass, "alias_nested_items", {}),
            dacite_config=getattr(klass, "dacite_config", None),
        )

    def get_alias_nested_items(self, name: str) -> typing.Optional[str]:
        return self.alias_nested_items.get(name)


@dataclasses.dataclass
class FieldMetadata:
    aliases: typing.List[str] = dataclasses.field(default_factory=list)
    doc: typing.Optional[str] = None
    namespace: typing.Optional[str] = None

    @classmethod
    def create(cls: typing.Type["FieldMetadata"], klass: typing.Optional[type]) -> "FieldMetadata":
        return cls(
            aliases=getattr(klass, "aliases", []),
            doc=getattr(klass, "doc", None),
            namespace=getattr(klass, "namespace", None),
        )

    def to_dict(self) -> typing.Dict[str, typing.Union[typing.List[str], str]]:
        return {key: value for key, value in vars(self).items() if value}


class UserDefinedType(typing.NamedTuple):
    name: str
    type: typing.Any


epoch: datetime = datetime(1970, 1, 1, tzinfo=utc)
epoch_naive: datetime = datetime(1970, 1, 1)



import datetime
import decimal
import uuid

import pytest

from dataclasses_avroschema import field_utils, fields, types

from . import consts


@pytest.mark.parametrize("python_type,avro_type,logical_type", consts.LOGICAL_TYPES_AND_DEFAULTS)
def test_logical_types(python_type, avro_type, logical_type):
    name = "a logical type"
    python_type = python_type
    field = fields.AvroField(name, python_type)

    expected = {"name": name, "type": {"type": avro_type, "logicalType": logical_type}}

    assert expected == field.to_dict()


@pytest.mark.parametrize("python_type,avro_type,logical_type", consts.LOGICAL_TYPES_AND_DEFAULTS)
def test_logical_types_with_null_as_default(python_type, avro_type, logical_type):
    name = "a logical type"
    python_type = python_type
    field = fields.AvroField(name, python_type, default=None)

    expected = {
        "name": name,
        "type": ["null", {"type": avro_type, "logicalType": logical_type}],
        "default": None,
    }

    assert expected == field.to_dict()


def test_logical_type_date_with_default():
    name = "a date"
    python_type = datetime.date
    field = fields.AvroField(name, python_type, default=consts.now.date())

    date_time = datetime.datetime.combine(consts.now, datetime.datetime.min.time())
    ts = (date_time - datetime.datetime(1970, 1, 1)).total_seconds()

    expected = {
        "name": name,
        "type": {"type": field_utils.INT, "logicalType": field_utils.DATE},
        "default": ts / (3600 * 24),
    }

    assert expected == field.to_dict()


def test_logical_type_time_millis_with_default() -> None:
    name = "a time"
    python_type = datetime.time
    time = consts.now.time()
    field = fields.AvroField(name, python_type, default=time)

    hour, minutes, seconds, microseconds = (
        time.hour,
        time.minute,
        time.second,
        time.microsecond,
    )
    miliseconds = int((((hour * 60 + minutes) * 60 + seconds) * 1000) + (microseconds / 1000))
    expected = {
        "name": name,
        "type": {"type": field_utils.INT, "logicalType": field_utils.TIME_MILLIS},
        "default": miliseconds,
    }

    assert expected == field.to_dict()


def test_logical_type_time_micros_with_default() -> None:
    name = "a time"
    python_type = types.TimeMicro
    time = consts.now.time()
    field = fields.AvroField(name, python_type, default=time)

    hour, minutes, seconds, microseconds = (
        time.hour,
        time.minute,
        time.second,
        time.microsecond,
    )
    microseconds = float(((hour * 60 + minutes) * 60 + seconds) * 1000000) + microseconds

    expected = {
        "name": name,
        "type": {"type": field_utils.LONG, "logicalType": field_utils.TIME_MICROS},
        "default": microseconds,
    }

    print(field, "marcos;")
    assert expected == field.to_dict()


def test_logical_type_datetime_micros_with_default() -> None:
    name = "a datetime"
    python_type = types.DateTimeMicro
    field = fields.AvroField(name, python_type, default=consts.now)

    ts = (consts.now - datetime.datetime(1970, 1, 1)).total_seconds()

    expected = {
        "name": name,
        "type": {"type": field_utils.LONG, "logicalType": field_utils.TIMESTAMP_MICROS},
        "default": ts * 1000000,
    }

    assert expected == field.to_dict()


def test_logical_type_datetime_with_default() -> None:
    name = "a datetime"
    python_type = datetime.datetime
    field = fields.AvroField(name, python_type, default=consts.now)

    ts = (consts.now - datetime.datetime(1970, 1, 1)).total_seconds()

    expected = {
        "name": name,
        "type": {"type": field_utils.LONG, "logicalType": field_utils.TIMESTAMP_MILLIS},
        "default": int(ts * 1000),
    }

    assert expected == field.to_dict()


@pytest.mark.parametrize(
    "python_type",
    (
        uuid.uuid4,
        uuid.UUID,
    ),
)
def test_logical_type_uuid_with_default(python_type) -> None:
    name = "a uuid"
    default = uuid.UUID("d793fc4f-2eef-440a-af8b-a8e884d7b1a8")
    field = fields.AvroField(name, python_type, default=default)

    expected = {
        "name": name,
        "type": {"type": field_utils.STRING, "logicalType": field_utils.UUID},
        "default": str(default),
    }

    assert expected == field.to_dict()


@pytest.mark.parametrize("logical_type,invalid_default,msg", consts.LOGICAL_TYPES_AND_INVALID_DEFAULTS)
def test_invalid_default_values(logical_type, invalid_default, msg):
    name = "a_field"
    field = fields.AvroField(name, logical_type, default=invalid_default)

    msg = msg or f"Invalid default type. Default should be {logical_type}"
    with pytest.raises(AssertionError, match=msg):
        field.to_dict()


def test_decimal_type():
    """
    When the type is types.Decimal, the Avro field type should be bytes,
    with logicalType=decimal and metadata attributes scale, precision present as ints
    """
    name = "a_decimal_field"
    # A default decimal.Decimal sets precision and scale implicitly
    default = decimal.Decimal("3.14")
    python_type = decimal.Decimal
    field = fields.AvroField(name, python_type, default=default)

    expected = {
        "name": name,
        "type": {
            "type": "bytes",
            "logicalType": "decimal",
            "precision": 3,
            "scale": 2,
        },
        "default": "\\u013a",
    }

    assert expected == field.to_dict()

    # Use types.Decimal to set explicitly
    default = types.Decimal(scale=5, precision=7)
    field = fields.AvroField(name, python_type, default=default)

    expected = {
        "name": name,
        "type": {
            "type": "bytes",
            "logicalType": "decimal",
            "precision": 7,
            "scale": 5,
        },
    }

    assert expected == field.to_dict()

    default = types.Decimal(scale=5, precision=7, default=decimal.Decimal("3.14"))
    field = fields.AvroField(name, python_type, default=default)

    expected = {
        "name": name,
        "type": {
            "type": "bytes",
            "logicalType": "decimal",
            "precision": 7,
            "scale": 5,
        },
        "default": "\\u04ca90",
    }

    assert expected == field.to_dict()

    default = types.Decimal(scale=5, precision=7, default=None)
    field = fields.AvroField(name, python_type, default=default)

    expected = {
        "name": name,
        "type": ["null", {"type": "bytes", "logicalType": "decimal", "precision": 7, "scale": 5}],
        "default": None,
    }

    assert expected == field.to_dict()

    # If default is missing, default out scale by Avro spec and pull precision from default decimal context
    # On my machine, this makes the "decimal" field a glorified 28-digit int, which is likely not what is wanted
    # so there is a good argument to error this out and force the dev to provide a default
    # default = types.MissingSentinel
    # field = fields.AvroField(name, python_type, default)
    #
    # expected = {
    #     "name": name,
    #     "type": {
    #         "type": "bytes",
    #         "logicalType": "decimal",
    #         "precision": decimal.Context().prec,
    #         "scale": 0,
    #     },
    # }
    #
    # assert expected == field.to_dict()

    # Validate 0 <= scale <= precision
    with pytest.raises(
        ValueError, match="Scale must be zero or a positive integer less than or equal to the precision."
    ):
        default = types.Decimal(scale=-1, precision=1)
        field = fields.AvroField(name, python_type, default=default)

        field.to_dict()

    # Validate 0 <= scale <= precision
    with pytest.raises(
        ValueError, match="Scale must be zero or a positive integer less than or equal to the precision."
    ):
        default = types.Decimal(scale=3, precision=1)
        field = fields.AvroField(name, python_type, default=default)

        field.to_dict()

    # Validate precision >= 0
    with pytest.raises(ValueError, match="Precision must be a positive integer greater than zero"):
        default = types.Decimal(scale=2, precision=-1)
        field = fields.AvroField(name, python_type, default=default)

        field.to_dict()

    # Require a default be provided for decimal.Decimal
    with pytest.raises(ValueError):
        default = types.MissingSentinel
        field = fields.AvroField(name, python_type, default=default)

        field.to_dict()

    # Catch unexpected default value for decimal.Decimal
    with pytest.raises(ValueError):
        default = 7
        field = fields.AvroField(name, python_type, default=default)

        field.to_dict()

    # Default decimal.Decimal has more digits than listed precision
    with pytest.raises(ValueError):
        default = types.Decimal(scale=2, precision=3, default=decimal.Decimal("3.14159"))
        field = fields.AvroField(name, python_type, default=default)

        field.to_dict()

    # Default decimal.Decimal has more digits past decimal than scale
    with pytest.raises(ValueError):
        default = types.Decimal(scale=1, precision=3, default=decimal.Decimal("3.14"))
        field = fields.AvroField(name, python_type, default=default)

        field.to_dict()

    # Just for code coverage
    with pytest.raises(ValueError):
        default = types.Decimal(scale=1, precision=3, default=decimal.Decimal("3.14"))
        field = fields.AvroField(name, python_type, default=default)

        field.to_dict()



from string import Template

FIELD_TEMPLATE = "$name: $type"
METACLASS_FIELD_TEMPLATE = '$name = "$value"'
METACLASS_ALIAS_FIELD = "$name = $value"
FIELD_DEFAULT_TEMPLATE = " = $default"
OPTIONAL_TEMPLATE = "typing.Optional[$type] = None"
UNION_TEMPLATE = "typing.Union[$type]"
LIST_TEMPLATE = "typing.List[$type]"
DICT_TEMPLATE = "typing.Dict[str, $type]"
FIXED_TEMPLATE = "types.Fixed = types.Fixed($properties)"
DATACLASS_FIELD = "dataclasses.field($properties)"
TYPE_TEMPLATE = 'typing.Type["$type"]'
DATE_TEMPLATE = "datetime.date($year, $month, $day)"
TIME_TEMPLATE = "datetime.time($hour, $minute, $second)"
TIME_MICROS_TEMPLATE = "datetime.time($hour, $minute, $second, $microsecond)"
DATETIME_TEMPLATE = "datetime.datetime($year, $month, $day, $hour, $minute, $second)"
DATETIME_MICROS_TEMPLATE = "datetime.datetime($year, $month, $day, $hour, $minute, $second, $microsecond)"
DECIMAL_TEMPLATE = "decimal.Decimal('$value')"
DECIMAL_TYPE_TEMPLATE = "types.Decimal($properties)"

ENUM_SYMBOL_TEMPLATE = "$key = $value"
ENUM_TEMPLATE = """

class $name(enum.Enum):
    $symbols
"""

CLASS_TEMPLATE = """
$decorator
class $name($base_class):
    $fields
"""

METACLASS_TEMPLATE = """
class Meta:
    $properties
"""

IMPORTS_TEMPLATE = """
$imports
"""

MODULE_TEMPLATE = """
$imports
$extras
$classes
"""

# Pydanntic specific
PYDANTIC_FIELD = "Field($properties)"
PYDANTIC_DECIMAL_TYPE_TEMPLATE = "condecimal(max_digits=$precision, decimal_places=$scale)"

field_template = Template(FIELD_TEMPLATE)
metaclass_field_template = Template(METACLASS_FIELD_TEMPLATE)
metaclass_alias_field_template = Template(METACLASS_ALIAS_FIELD)
field_default_template = Template(FIELD_DEFAULT_TEMPLATE)
optional_template = Template(OPTIONAL_TEMPLATE)
union_template = Template(UNION_TEMPLATE)
list_template = Template(LIST_TEMPLATE)
dict_template = Template(DICT_TEMPLATE)
fixed_template = Template(FIXED_TEMPLATE)
decimal_template = Template(DECIMAL_TEMPLATE)
decimal_type_template = Template(DECIMAL_TYPE_TEMPLATE)
pydantic_decimal_type_template = Template(PYDANTIC_DECIMAL_TYPE_TEMPLATE)
enum_symbol_template = Template(ENUM_SYMBOL_TEMPLATE)
enum_template = Template(ENUM_TEMPLATE)
dataclass_field_template = Template(DATACLASS_FIELD)
pydantic_field_template = Template(PYDANTIC_FIELD)
class_template = Template(CLASS_TEMPLATE)
metaclass_template = Template(METACLASS_TEMPLATE)
type_template = Template(TYPE_TEMPLATE)
date_template = Template(DATE_TEMPLATE)
time_template = Template(TIME_TEMPLATE)
time_micros_template = Template(TIME_MICROS_TEMPLATE)
datetime_template = Template(DATETIME_TEMPLATE)
datetime_micros_template = Template(DATETIME_MICROS_TEMPLATE)
imports_template = Template(IMPORTS_TEMPLATE.strip())
module_template = Template(MODULE_TEMPLATE.strip())



import dataclasses
import enum
import inspect
import json
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Set, Type, TypeVar, Union

from dacite import Config, from_dict
from fastavro.validation import validate

from . import case
from .fields import EnumField, FieldType, RecordField, UnionField
from .schema_definition import AvroSchemaDefinition
from .serialization import deserialize, serialize, to_json
from .types import Decimal, Fixed, JsonDict
from .utils import SchemaMetadata, is_dataclass_or_pydantic_model

AVRO = "avro"
AVRO_JSON = "avro-json"


CT = TypeVar("CT", bound="AvroModel")


class AvroModel:
    schema_def: Optional[AvroSchemaDefinition] = None
    klass: Optional[Type] = None
    metadata: Optional[SchemaMetadata] = None
    user_defined_types: Set = set()
    parent: Any = None
    rendered_schema: OrderedDict = dataclasses.field(default_factory=OrderedDict)

    @classmethod
    def generate_dataclass(cls: Type[CT]) -> Type[CT]:
        if is_dataclass_or_pydantic_model(cls):
            return cls
        return dataclasses.dataclass(cls)

    @classmethod
    def generate_metadata(cls: Type[CT]) -> SchemaMetadata:
        meta = getattr(cls.klass, "Meta", type)

        return SchemaMetadata.create(meta)

    @classmethod
    def generate_schema(cls: Type[CT], schema_type: str = "avro") -> Optional[OrderedDict]:
        if cls.schema_def is None or cls.__mro__[1] != AvroModel:
            # Generate dataclass and metadata
            cls.klass = cls.generate_dataclass()

            # let's live open the possibility to define different
            # schema definitions like json
            if schema_type == "avro":
                # cache the schema
                cls.schema_def = cls._generate_avro_schema()
                cls.rendered_schema = cls.schema_def.render()
            else:
                raise ValueError("Invalid type. Expected avro schema type.")

        return cls.rendered_schema

    @classmethod
    def _generate_avro_schema(cls: Type[CT]) -> AvroSchemaDefinition:
        metadata = cls.generate_metadata()
        cls.metadata = metadata
        return AvroSchemaDefinition("record", cls.klass, metadata=metadata, parent=cls.parent or cls)

    @classmethod
    def avro_schema(cls: Type[CT], case_type: Optional[str] = None) -> str:
        avro_schema = cls.generate_schema(schema_type=AVRO)

        if case_type is not None:
            avro_schema = case.case_record(cls.rendered_schema, case_type)  # type: ignore

        return json.dumps(avro_schema)

    @classmethod
    def avro_schema_to_python(cls: Type[CT], parent: Optional["AvroModel"] = None) -> Dict[str, Any]:
        if parent is not None:
            # in this case the current class is a child with a parent
            # we recalculate the schema definition to prevent re usages
            cls.parent = parent
            cls.schema_def = None
        else:
            # This happens when an AvroModel is the root of the tree (first class in the hierarchy)
            # Because intermediate schemas can be reused as a root later, we need to reset them
            # Example with A as a root:
            #     A -> B -> C -> D
            #
            # After generating the A.avro_schema the parent of B is A,
            # if we want to do B.avro_schema (now B is the root)
            # B should clean the data that was only valid when it was the child
            cls._reset_schema_definition()

        return json.loads(cls.avro_schema())

    @classmethod
    def get_fields(cls: Type[CT]) -> List[FieldType]:
        if cls.schema_def is None:
            cls.generate_schema()
        return cls.schema_def.fields  # type: ignore

    @classmethod
    def _get_enum_type_map(cls: Type[CT]) -> Dict[str, enum.EnumMeta]:
        enum_types = {}
        for field_type in cls.get_fields():
            if isinstance(field_type, EnumField):
                enum_types[field_type.name] = field_type.type
            elif isinstance(field_type, UnionField):
                for sub_type in field_type.type.__args__:
                    if inspect.isclass(sub_type) and issubclass(sub_type, enum.Enum):
                        enum_types[field_type.name] = sub_type
            elif isinstance(field_type, RecordField):
                enum_types.update(field_type.type._get_enum_type_map())
        return enum_types

    @classmethod
    def _deserialize_complex_types(cls: Type[CT], payload: Dict[str, Any]) -> Dict:
        output = {}
        enum_type_map = cls._get_enum_type_map()
        for field, value in payload.items():
            if isinstance(value, dict):
                output[field] = cls._deserialize_complex_types(value)
            elif field in enum_type_map and isinstance(value, str):
                try:
                    enum_field = enum_type_map[field]
                    output[field] = enum_field(value)
                except ValueError as e:
                    raise ValueError(f"Value {value} is not a valid instance of {enum_type_map[field]}", e)
            else:
                output[field] = value
        return output

    @classmethod
    def _reset_schema_definition(cls: Type[CT]) -> None:
        """
        Reset all the values to original state.
        """
        cls.user_defined_types = set()
        cls.schema_def = None
        cls.parent = None

    @staticmethod
    def standardize_custom_type(value: Any) -> Any:
        if isinstance(value, (Decimal, Fixed)):
            return value.default
        elif isinstance(value, dict):
            return {k: AvroModel.standardize_custom_type(v) for k, v in value.items()}
        elif isinstance(value, (list, tuple)):
            return [AvroModel.standardize_custom_type(v) for v in value]
        elif issubclass(type(value), enum.Enum):
            return value.value
        return value

    def asdict(self) -> JsonDict:
        return dataclasses.asdict(
            self, dict_factory=lambda x: {key: self.standardize_custom_type(value) for key, value in x}
        )

    def serialize(self, serialization_type: str = AVRO) -> bytes:
        schema = self.avro_schema_to_python()

        return serialize(self.asdict(), schema, serialization_type=serialization_type)

    @classmethod
    def deserialize(
        cls: Type[CT],
        data: bytes,
        serialization_type: str = AVRO,
        create_instance: bool = True,
        writer_schema: Optional[Union[JsonDict, Type[CT]]] = None,
    ) -> Union[JsonDict, CT]:
        if inspect.isclass(writer_schema) and issubclass(writer_schema, AvroModel):
            # mypy does not undersdtand redefinitions
            writer_schema: JsonDict = writer_schema.avro_schema_to_python()  # type: ignore

        schema = cls.avro_schema_to_python()
        payload = deserialize(
            data, schema, serialization_type=serialization_type, writer_schema=writer_schema  # type: ignore
        )
        output = cls._deserialize_complex_types(payload)

        if create_instance:
            return cls.parse_obj(data=output)
        return output

    @classmethod
    def parse_obj(cls: Type[CT], data: Dict) -> Union[JsonDict, CT]:
        return from_dict(data_class=cls, data=data, config=Config(**cls.config()))

    def validate(self) -> bool:
        schema = self.avro_schema_to_python()
        return validate(self.asdict(), schema)

    def to_dict(self) -> JsonDict:
        # Serialize using the current AVRO schema to get proper field representations
        # and after that convert into python
        return self.asdict()

    def to_json(self) -> str:
        data = to_json(self.asdict())
        return json.dumps(data)

    @classmethod
    def config(cls: Type[CT]) -> JsonDict:
        """
        Get the default config for dacite and always include the self reference
        """
        # We need to make sure that the `avro schemas` has been generated, otherwise cls.klass is empty
        # It won't affect the performance because the rendered schema will be store in cls.rendered_schema
        cls.generate_schema()
        dacite_user_config = cls.metadata.dacite_config  # type: ignore

        dacite_config = {
            "check_types": False,
            "forward_references": {
                cls.klass.__name__: cls.klass,  # type: ignore
            },
        }

        if dacite_user_config is not None:
            dacite_config.update(dacite_user_config)

        return dacite_config

    @classmethod
    def fake(cls: Type[CT], **data: Dict[str, Any]) -> CT:
        """
        Creates a fake instance of the model.

        Attributes:
            data: Dict[str, Any] represent the user values to use in the instance
        """
        # only generate fakes for fields that were not provided in data
        payload = {field.name: field.fake() for field in cls.get_fields() if field.name not in data.keys()}
        payload.update(data)

        return from_dict(data_class=cls, data=payload, config=Config(**cls.config()))



import datetime
import sys
import typing
import uuid

import pytest

from dataclasses_avroschema import field_utils

PY_VER = sys.version_info

now = datetime.datetime.now()

PRIMITIVE_TYPES = (
    (str, field_utils.STRING),
    (int, field_utils.LONG),
    (bool, field_utils.BOOLEAN),
    (float, field_utils.DOUBLE),
    (bytes, field_utils.BYTES),
)

PRIMITIVE_TYPES_AND_DEFAULTS = (
    (str, "test"),
    (int, 1),
    (bool, True),
    (float, 10.4),
    (bytes, b"test"),
)

PRIMITIVE_TYPES_AND_INVALID_DEFAULTS = (
    (str, 1),
    (int, "test"),
    (bool, 10),
    (float, False),
    (bytes, "test"),
)

LIST_TYPE_AND_ITEMS_TYPE = (
    (str, "string"),
    (int, "long"),
    (bool, "boolean"),
    (float, "double"),
    (bytes, "bytes"),
)

LOGICAL_TYPES = (
    (datetime.date, field_utils.LOGICAL_DATE, now.date()),
    (datetime.time, field_utils.LOGICAL_TIME_MILIS, now.time()),
    (datetime.datetime, field_utils.LOGICAL_DATETIME_MILIS, now),
    (uuid.uuid4, field_utils.LOGICAL_UUID, uuid.uuid4()),
)

UNION_PRIMITIVE_ELEMENTS = (
    (
        typing.Union[str, int],
        (field_utils.STRING, field_utils.LONG),
    ),
    (
        typing.Union[str, bytes],
        (field_utils.STRING, field_utils.BYTES),
    ),
    (
        typing.Union[str, None],
        (field_utils.STRING, field_utils.NULL),
    ),
    (
        typing.Union[datetime.date, datetime.datetime],
        (
            field_utils.PYTHON_TYPE_TO_AVRO[datetime.date],
            field_utils.PYTHON_TYPE_TO_AVRO[datetime.datetime],
        ),
    ),
    (typing.Union[float, str, int], (field_utils.DOUBLE, field_utils.STRING, field_utils.LONG)),
    (
        typing.Union[str, float, int, bool],
        (field_utils.STRING, field_utils.DOUBLE, field_utils.LONG, field_utils.BOOLEAN),
    ),
)

UNION_PRIMITIVE_ELEMENTS_DEFAULTS = (
    (typing.Union[str, int], (field_utils.STRING, field_utils.LONG), "test"),
    (typing.Union[str, bytes], (field_utils.BYTES, field_utils.STRING), b"test"),
    (typing.Union[str, None], (field_utils.NULL, field_utils.STRING), None),
    (
        typing.Union[datetime.date, datetime.datetime],
        (
            field_utils.PYTHON_TYPE_TO_AVRO[datetime.datetime],
            field_utils.PYTHON_TYPE_TO_AVRO[datetime.date],
        ),
        now,
    ),
    (typing.Union[float, str, int], (field_utils.DOUBLE, field_utils.STRING, field_utils.LONG), 100.0),
    (
        typing.Union[str, float, int, bool],
        (field_utils.BOOLEAN, field_utils.STRING, field_utils.DOUBLE, field_utils.LONG),
        False,
    ),
)

UNION_WITH_ARRAY = (
    (
        typing.Union[typing.List[int], str],
        (field_utils.LONG, field_utils.STRING),
    ),
    (
        typing.Union[typing.List[str], float],
        (field_utils.STRING, field_utils.DOUBLE),
    ),
    (
        typing.Union[typing.List[datetime.datetime], datetime.datetime],
        (field_utils.LOGICAL_DATETIME_MILIS, field_utils.LOGICAL_DATETIME_MILIS),
    ),
    (
        typing.Union[typing.List[uuid.uuid4], bytes],
        (field_utils.LOGICAL_UUID, field_utils.BYTES),
    ),
)

UNION_WITH_MAP = (
    (
        typing.Union[typing.Dict[str, int], str],
        (field_utils.LONG, field_utils.STRING),
    ),
    (
        typing.Union[typing.Dict[str, str], float],
        (field_utils.STRING, field_utils.DOUBLE),
    ),
    (
        typing.Union[typing.Dict[str, datetime.datetime], datetime.datetime],
        (field_utils.LOGICAL_DATETIME_MILIS, field_utils.LOGICAL_DATETIME_MILIS),
    ),
    (
        typing.Union[typing.Dict[str, uuid.uuid4], bytes],
        (field_utils.LOGICAL_UUID, field_utils.BYTES),
    ),
)

OPTIONAL_UNION_COMPLEX_TYPES = (
    (typing.List[str], {"type": field_utils.ARRAY, "items": field_utils.STRING, "name": "optional_field"}),
    (
        typing.List[datetime.datetime],
        {"type": field_utils.ARRAY, "items": field_utils.LOGICAL_DATETIME_MILIS, "name": "optional_field"},
    ),
    (typing.Dict[str, int], {"type": field_utils.MAP, "values": field_utils.LONG, "name": "optional_field"}),
    (
        typing.Dict[str, datetime.datetime],
        {"type": field_utils.MAP, "values": field_utils.LOGICAL_DATETIME_MILIS, "name": "optional_field"},
    ),
)

# Add the new syntax | for unions
if PY_VER >= (3, 10):
    UNION_PRIMITIVE_ELEMENTS += (  # type: ignore
        (
            str | int,
            (field_utils.STRING, field_utils.LONG),
        ),
        (
            str | bytes,
            (field_utils.STRING, field_utils.BYTES),
        ),
        (
            str | None,
            (field_utils.STRING, field_utils.NULL),
        ),
        (
            datetime.date | datetime.datetime,
            (
                field_utils.PYTHON_TYPE_TO_AVRO[datetime.date],
                field_utils.PYTHON_TYPE_TO_AVRO[datetime.datetime],
            ),
        ),
        (float | str | int, (field_utils.DOUBLE, field_utils.STRING, field_utils.LONG)),
        (
            typing.Union[str, float, int, bool],
            (field_utils.STRING, field_utils.DOUBLE, field_utils.LONG, field_utils.BOOLEAN),
        ),
    )

    UNION_PRIMITIVE_ELEMENTS_DEFAULTS += (  # type: ignore
        (str | int, (field_utils.STRING, field_utils.LONG), "test"),
        (str | bytes, (field_utils.BYTES, field_utils.STRING), b"test"),
        (str | None, (field_utils.NULL, field_utils.STRING), None),
        (
            datetime.date | datetime.datetime,
            (
                field_utils.PYTHON_TYPE_TO_AVRO[datetime.datetime],
                field_utils.PYTHON_TYPE_TO_AVRO[datetime.date],
            ),
            now,
        ),
        (float | str | int, (field_utils.DOUBLE, field_utils.STRING, field_utils.LONG), 100.0),
        (
            str | float | int | bool,
            (field_utils.BOOLEAN, field_utils.STRING, field_utils.DOUBLE, field_utils.LONG),
            False,
        ),
    )

    UNION_WITH_ARRAY += (  # type: ignore
        (
            typing.List[int] | str,  # in python is translated to typing.Union[typing.List[int], str]
            (field_utils.LONG, field_utils.STRING),
        ),
        (
            typing.List[str] | float,
            (field_utils.STRING, field_utils.DOUBLE),
        ),
        (
            typing.List[datetime.datetime] | datetime.datetime,
            (field_utils.LOGICAL_DATETIME_MILIS, field_utils.LOGICAL_DATETIME_MILIS),
        ),
        (
            typing.List[uuid.UUID] | bytes,
            (field_utils.LOGICAL_UUID, field_utils.BYTES),
        ),
    )

    UNION_WITH_MAP += (  # type: ignore
        (
            typing.Dict[str, int] | str,  # in python is translated to typing.Union[typing.Dict[str, int], str]
            (field_utils.LONG, field_utils.STRING),
        ),
        (
            typing.Dict[str, str] | float,
            (field_utils.STRING, field_utils.DOUBLE),
        ),
        (
            typing.Dict[str, datetime.datetime] | datetime.datetime,
            (field_utils.LOGICAL_DATETIME_MILIS, field_utils.LOGICAL_DATETIME_MILIS),
        ),
        (
            typing.Dict[str, uuid.UUID] | bytes,
            (field_utils.LOGICAL_UUID, field_utils.BYTES),
        ),
    )


def xfail_annotation(typ):
    return pytest.mark.xfail(
        (typ in (list, tuple, dict) and PY_VER < (3, 9)),
        reason="Standard collection annotations not supported - see PEP585.",
    )


SEQUENCE_TYPES = (typing.List, typing.Tuple, typing.Sequence, typing.MutableSequence, list, tuple)
MAPPING_TYPES = (typing.Dict, typing.Mapping, typing.MutableMapping, dict)

SEQUENCES_AND_TYPES = [
    pytest.param(sequence, python_type, items_type, marks=xfail_annotation(sequence))
    for sequence in SEQUENCE_TYPES
    for python_type, items_type in PRIMITIVE_TYPES
]

SEQUENCES_LOGICAL_TYPES = [
    pytest.param(sequence, python_type, items_type, value, marks=xfail_annotation(sequence))
    for sequence in SEQUENCE_TYPES
    for python_type, items_type, value in LOGICAL_TYPES
]

MAPPING_AND_TYPES = [
    pytest.param(mapping, python_type, items_type, marks=xfail_annotation(mapping))
    for mapping in MAPPING_TYPES
    for python_type, items_type in PRIMITIVE_TYPES
]

MAPPING_LOGICAL_TYPES = [
    pytest.param(mapping, python_type, items_type, value, marks=xfail_annotation(mapping))
    for mapping in MAPPING_TYPES
    for python_type, items_type, value in LOGICAL_TYPES
]

# Represent the logical types
# (python_type, avro_internal_type, logical_type)
LOGICAL_TYPES_AND_DEFAULTS = (
    (datetime.date, field_utils.INT, field_utils.DATE),
    (datetime.time, field_utils.INT, field_utils.TIME_MILLIS),
    (datetime.datetime, field_utils.LONG, field_utils.TIMESTAMP_MILLIS),
    (uuid.uuid4, field_utils.STRING, field_utils.UUID),
    (uuid.UUID, field_utils.STRING, field_utils.UUID),
)

LOGICAL_TYPES_AND_INVALID_DEFAULTS = (
    (datetime.date, 1, None),
    (datetime.time, "test", None),
    (datetime.datetime, 10, None),
    (uuid.uuid4, 10, f"Invalid default type. Default should be {str} or {uuid.UUID}"),
    (uuid.UUID, 10, f"Invalid default type. Default should be {str} or {uuid.UUID}"),
)


ARRAY_WITH_UNION_TYPES = (
    (typing.Union[int, str], [field_utils.LONG, field_utils.STRING], [10, 20, "test"]),
    (
        typing.Union[int, str, float],
        [field_utils.LONG, field_utils.STRING, field_utils.DOUBLE],
        [10, 20, 100.9],
    ),
)



import typing
from dataclasses import dataclass, field
from string import Template

import fastavro
import stringcase

from dataclasses_avroschema import field_utils, serialization
from dataclasses_avroschema.types import JsonDict

from . import avro_to_python_utils, templates
from .base_class import BaseClassEnum


@dataclass
class ModelGenerator:
    base_class: str = BaseClassEnum.AVRO_MODEL.value
    field_identation: str = "\n    "
    imports: typing.Set[str] = field(default_factory=set)
    # extras is used fot extra code that is generated, for example Enums
    extras: typing.List[str] = field(default_factory=list)
    class_template: Template = templates.class_template
    dataclass_field_template: Template = templates.dataclass_field_template
    metadata_fields_mapper: typing.Dict[str, str] = field(
        default_factory=lambda: {
            "namespace": "namespace",
            "doc": "schema_doc",
            "alias": "aliases",
        }
    )
    matadata_field_templates: typing.Dict[str, Template] = field(
        default_factory=lambda: {
            "namespace": templates.metaclass_field_template,
            "doc": templates.metaclass_field_template,
            "alias": templates.metaclass_alias_field_template,
        }
    )
    base_class_to_imports: typing.Dict[str, str] = field(
        default_factory=lambda: {
            BaseClassEnum.AVRO_MODEL.value: "from dataclasses_avroschema import AvroModel",
            BaseClassEnum.PYDANTIC_MODEL.value: "from pydantic import BaseModel",
            BaseClassEnum.AVRO_DANTIC_MODEL.value: "from dataclasses_avroschema.avrodantic import AvroBaseModel",
        }
    )
    # represent the decorator to add in the base class
    base_class_decotator: str = ""
    avro_type_to_python: typing.Dict[str, str] = field(init=False)
    logical_types_imports: typing.Dict[str, str] = field(init=False)

    def __post_init__(self) -> None:
        self.imports.add(self.base_class_to_imports[self.base_class])
        self.avro_type_to_python = avro_to_python_utils.AVRO_TYPE_TO_PYTHON[self.base_class]
        self.logical_types_imports = avro_to_python_utils.LOGICAL_TYPES_IMPORTS[self.base_class]

        if self.base_class == BaseClassEnum.AVRO_MODEL.value:
            self.imports.add("import dataclasses")
            self.base_class_decotator = "@dataclasses.dataclass"
        else:
            self.dataclass_field_template = templates.pydantic_field_template

    @staticmethod
    def validate_schema(*, schema: JsonDict) -> None:
        """
        Validate that the schema is a valid avro schema
        """
        fastavro.parse_schema(schema)

    def render_imports(self) -> str:
        """
        Render the imports needed for the python classes.
        """
        # sort the imports
        list_imports = list(self.imports)
        list_imports.sort()

        imports = "\n".join([imp for imp in list_imports])
        return templates.imports_template.safe_substitute(imports=imports)

    def render_extras(self) -> str:
        return "".join([extra for extra in self.extras])

    def render_metaclass(self, *, schema: JsonDict) -> typing.Optional[str]:
        """
        Render Class Meta that contains the schema matadata
        """
        properties = self.field_identation.join(
            [
                self.matadata_field_templates[meta_avro_field].safe_substitute(
                    name=meta_field, value=schema.get(meta_avro_field)
                )
                for meta_avro_field, meta_field in self.metadata_fields_mapper.items()
                if schema.get(
                    meta_avro_field
                )  # TODO: replace this line with if (value := schema.get(meta_avro_field)) after drop py3.7
            ]
        )

        if properties:
            # some formating to remove identation at the end of the Class Meta to make it more compatible with black
            return (
                self.field_identation.join(
                    [line for line in templates.metaclass_template.safe_substitute(properties=properties).split("\n")]
                ).rstrip(self.field_identation)
                + "\n"
            )
        return None

    def render_class(self, *, schema: JsonDict) -> str:
        """
        Render the class generated from the schema
        """
        name: str = stringcase.pascalcase(schema["name"])
        record_fields: typing.List[JsonDict] = schema["fields"]

        # Sort the fields according whether it has a default value
        record_fields.sort(key=lambda field: 1 if "default" in field.keys() or field_utils.NULL in field["type"] else 0)

        rendered_fields = self.field_identation.join(
            [self.render_field(field=field, model_name=name) for field in record_fields]
        )

        rendered_class = templates.class_template.safe_substitute(
            name=name, decorator=self.base_class_decotator, base_class=self.base_class, fields=rendered_fields
        )

        class_metadata = self.render_metaclass(schema=schema)
        if class_metadata is not None:
            rendered_class += class_metadata

        return rendered_class

    def render_dataclass_field(self, properties: str) -> str:
        if self.base_class != BaseClassEnum.AVRO_MODEL.value:
            self.imports.add("from pydantic import Field")

        return self.dataclass_field_template.safe_substitute(properties=properties)

    def render(self, *, schema: JsonDict) -> str:
        """
        Render the module with the classes generated from the schema
        """
        return self.render_module(schemas=[schema])

    def render_module(self, *, schemas: typing.List[JsonDict]) -> str:
        """
        Render the module with the classes generated from the schemas
        """
        for schema in schemas:
            self.validate_schema(schema=schema)

        classes = "\n".join(self.render_class(schema=schema) for schema in schemas)
        imports = self.render_imports()
        extras = self.render_extras()

        return templates.module_template.safe_substitute(
            classes=classes,
            imports=imports,
            extras=extras,
        ).lstrip("\n")

    def render_field(self, field: JsonDict, model_name: str) -> str:
        """
        Render an avro field.

        1. If the field is a native one, we just render it.
        2. If the field is a record, array, map, fixed, enum
            we need to render the field again (recursive)
        3. If the field is a LogicalType, it may not have the
            the `name` property and the type is a `native` one
        """
        name = field.get("name", "")
        type = field["type"]
        default = field.get("default")

        # This flag tells whether the field is array, map, fixed
        is_complex_type = False

        if self.is_logical_type(field=field):
            is_complex_type = True
            # override the type so it can be use to get the default value in case that is needed
            type = field.get("logicalType") or field["type"]["logicalType"]
            language_type = self.parse_logical_type(field=field)

            if type == field_utils.DECIMAL:
                # set default to None because all the decimal has a default by design
                # and they are calculated in parse_decimal method
                default = None
        elif isinstance(type, dict):
            language_type = self.render_field(field=type, model_name=model_name)
        elif isinstance(type, list):
            language_type = self.parse_union(field_types=type, model_name=model_name)
        elif type == field_utils.ARRAY:
            is_complex_type = True
            language_type = self.parse_array(field=field, model_name=model_name)
        elif type == field_utils.MAP:
            is_complex_type = True
            language_type = self.parse_map(field=field, model_name=model_name)
        elif type == field_utils.ENUM:
            is_complex_type = True
            language_type = self.parse_enum(field=field)
        elif type == field_utils.FIXED:
            is_complex_type = True
            language_type = self.parse_fixed(field=field)
        elif type == field_utils.RECORD:
            record = f"\n{self.render_class(schema=field)}"
            is_complex_type = True
            self.extras.append(record)
            language_type = stringcase.pascalcase(field["name"])
        else:
            # Native field or Logical type using a native
            language_type = self.get_language_type(type=type)

        if is_complex_type:
            # If the field is a complext type, we need to return just the language_type
            result = language_type
        else:
            result = templates.field_template.safe_substitute(name=name, type=language_type)

        if default is not None:
            # optional field attribute
            default = self.get_field_default(field_type=type, default=field.get("default"), name=name)
            result += templates.field_default_template.safe_substitute(default=default)

        return result

    @staticmethod
    def is_logical_type(*, field: JsonDict) -> bool:
        if field.get("logicalType"):
            return True

        field_type = field["type"]
        return isinstance(field_type, dict) and field_type.get("logicalType") is not None

    def parse_logical_type(self, *, field: JsonDict) -> str:
        logical_type = field.get("logicalType") or field["type"]["logicalType"]

        # add the logical type import
        self.imports.add(self.logical_types_imports[logical_type])
        field_name = field.get("name")

        if field_name is not None:
            field_repr = templates.field_template.safe_substitute(
                name=field_name, type=self.avro_type_to_python[logical_type]
            )

            if logical_type == field_utils.DECIMAL:
                # this is a special case for logical types
                default_decimal = self.parse_decimal(field=field)
                field_repr += templates.field_default_template.safe_substitute(default=default_decimal)

            return field_repr
        return self.avro_type_to_python[logical_type]

    def parse_decimal(self, *, field: JsonDict) -> str:
        schema: JsonDict = field["type"]
        precision = schema["precision"]
        scale = schema["scale"]
        properties = f"scale={scale}, precision={precision}"
        default = field.get("default")

        if self.base_class == BaseClassEnum.AVRO_MODEL.value:
            self.imports.add("from dataclasses_avroschema import types")

            if default is not None:
                default = templates.decimal_template.safe_substitute(
                    value=serialization.string_to_decimal(value=default, schema=schema)
                )
                properties += f", default={default}"
            return templates.decimal_type_template.safe_substitute(properties=properties)
        else:
            # we should render a pydantic condecimal
            self.imports.add("from pydantic import condecimal")
            return templates.pydantic_decimal_type_template.safe_substitute(scale=scale, precision=precision)

    def parse_union(self, *, field_types: typing.List, model_name: str) -> str:
        """
        Parse an Avro union

        An union field is an array like ["null", "str", "int", ...]

        If the first element is `null` it means that the property is optional.
        The first item is the actual type of the field a AVRO specifies.

        Attributes:
            field_types: List of avro types
            model_name: name of the model that contains the `union`
        """
        first_type = field_types[0]

        if first_type == field_utils.NULL:
            # It is an optional field, we should include in the imports typing
            # an use the optional Template
            self.imports.add("import typing")

            # the union can have native types or custom types
            # if the type is not in the AVRO_TYPE_TO_PYTHON it means that it is a CustomType
            if len(field_types) > 2:
                language_types = self.parse_union(field_types=field_types[1:], model_name=model_name)
            elif isinstance(field_types[1], dict):
                # TODO: check what happens with more than 2 complex types
                language_types = self.render_field(field=field_types[1], model_name=model_name)
            else:
                language_types = ", ".join(
                    # TODO: check default, seems doing nothing
                    [self.get_language_type(type=type, default=type, model_name=model_name) for type in field_types[1:]]
                )

            return templates.optional_template.safe_substitute(type=language_types)
        else:
            # It is a field that does not contain `null`
            # unions like ["string", "null"] don't make sense for the language as
            # the default value should be `string`, `null` is not possible as default

            # remove all the `null` apparences
            field_types = [field for field in field_types if field != field_utils.NULL]

            if len(field_types) >= 2:
                # an union with more than 2 types
                self.imports.add("import typing")

                language_types_list = []
                for type in field_types:
                    if isinstance(type, dict):
                        # it is a complex type like array, dict, enum, fixed or record
                        # it needs to be render
                        language_types_list.append(self.render_field(field=type, model_name=model_name))
                    else:
                        language_types_list.append(self.get_language_type(type=type, default=type))

                language_types_repr = ", ".join(
                    [self.get_language_type(type=type, default=type) for type in language_types_list]
                )
                return templates.union_template.safe_substitute(type=language_types_repr)
            else:
                # If for some reason `null` was inluded we should ignore it
                # the union has 2 types, so we return the first language type
                # example: "types": ["string", "null"]
                return self.get_language_type(type=first_type)

    def parse_array(self, field: JsonDict, model_name: str) -> str:
        """
        Parse an Avro array

        The type is is specify in the `items` attribute

        Example:
            {"name": "pets", "type": {"type": "array", "items": "string", "name": "pet"}}
        """
        type = field["items"]
        language_type = self._get_complex_langauge_type(type=type, model_name=model_name)

        return templates.list_template.safe_substitute(type=language_type)

    def parse_map(self, field: JsonDict, model_name: str) -> str:
        """
        Parse an Avro map

        The type is is specify in the `values` attribute. All the keys must be `string`

        Example:
            {"name": "accounts_money", "type": {"type": "map", "values": "float", "name": "accounts_money"}},
        """
        type = field["values"]
        language_type = self._get_complex_langauge_type(type=type, model_name=model_name)

        return templates.dict_template.safe_substitute(type=language_type)

    def parse_fixed(self, field: JsonDict) -> str:
        self.imports.add("from dataclasses_avroschema import types")
        properties = f"{field['size']}"

        namespace = field.get("namespace")
        aliases = field.get("aliases")

        if namespace is not None:
            properties += f', namespace="{namespace}"'

        if aliases is not None:
            properties += f", aliases={aliases}"

        return templates.fixed_template.safe_substitute(properties=properties)

    def parse_enum(self, field: JsonDict) -> str:
        """
        Parse an Avro Enum

        Avro Enums are asociated with enum.Enum types in python.
        We need a template for it in order to create the Enum
        """
        self.imports.add("import enum")

        field_name: str = field["name"]
        enum_name = stringcase.pascalcase(field_name)
        symbols = self.field_identation.join(
            [
                templates.enum_symbol_template.safe_substitute(key=stringcase.uppercase(symbol), value=f'"{symbol}"')
                for symbol in field["symbols"]
            ]
        )

        enum_class = templates.enum_template.safe_substitute(name=enum_name, symbols=symbols)
        self.extras.append(enum_class)

        return enum_name

    def _get_complex_langauge_type(self, *, type: typing.Any, model_name: str) -> str:
        """
        Get the language type for complext types (array and maps)
        """
        self.imports.add("import typing")

        if isinstance(type, dict):
            language_type = self.render_field(field=type, model_name=model_name)
        elif isinstance(type, list):
            language_type = self.parse_union(field_types=type, model_name=model_name)
        else:
            language_type = self.get_language_type(type=type, model_name=model_name)

        return language_type

    def get_language_type(
        self, *, type: str, default: typing.Optional[str] = None, model_name: typing.Optional[str] = None
    ) -> str:
        if type in (field_utils.FIXED, field_utils.INT, field_utils.FLOAT):
            self.imports.add("from dataclasses_avroschema import types")

        if type == model_name:
            # it means that it is a one-to-self-relationship
            return templates.type_template.safe_substitute(type=type)

        if default is not None:
            return str(self.avro_type_to_python.get(type, default))
        return str(self.avro_type_to_python.get(type, type))

    def get_field_default(self, *, field_type: str, default: typing.Any, name: str) -> typing.Any:
        """
        Returns the default value according to the field type

        TODO: docstrings

        Example:
            If the default is "bond" the method should return '\n"bond\n"' so the double quotes
            won't be scaped during the field render
        """
        if field_type in (field_utils.STRING, field_utils.UUID):
            return f'"{default}"'
        elif field_type == field_utils.BYTES:
            return f'b"{default}"'
        elif isinstance(field_type, dict) and field_type.get("type") == field_utils.ENUM:
            return f"{stringcase.pascalcase(name)}.{stringcase.uppercase(default)}"
        elif isinstance(
            default,
            (
                list,
                dict,
            ),
        ):
            if default:
                # it is an array or maps with some defaults that we should
                # express with a lambda function
                properties = f"default_factory=lambda: {default}"
            else:
                # it is an array or maps with `[]` or `{} ` as default
                properties = f"default_factory={default.__class__.__name__}"
            return self.render_dataclass_field(properties=properties)
        elif isinstance(field_type, list):
            return self.get_field_default(field_type=field_type[0], default=default, name=name)
        elif field_type in avro_to_python_utils.LOGICAL_TYPES_TO_PYTHON:
            func = avro_to_python_utils.LOGICAL_TYPES_TO_PYTHON[field_type]
            python_type = func(default)
            template_func = avro_to_python_utils.LOGICAL_TYPE_TEMPLATES[field_type]
            return template_func(python_type)
        else:
            return default



import datetime
import typing

from dataclasses_avroschema import field_utils

from . import templates
from .base_class import BaseClassEnum

_AVRO_TYPE_TO_PYTHON: typing.Dict[str, str] = {
    field_utils.BOOLEAN: "bool",
    field_utils.LONG: "int",
    field_utils.DOUBLE: "float",
    field_utils.BYTES: "bytes",
    field_utils.STRING: "str",
    field_utils.INT: "types.Int32",
    field_utils.FLOAT: "types.Float32",
    field_utils.DECIMAL: "decimal.Decimal",
    field_utils.DATE: "datetime.date",
    field_utils.TIME_MILLIS: "datetime.time",
    field_utils.TIME_MICROS: "types.TimeMicro",
    field_utils.TIMESTAMP_MILLIS: "datetime.datetime",
    field_utils.TIMESTAMP_MICROS: "types.DateTimeMicro",
    field_utils.UUID: "uuid.UUID",
}

AVRO_TYPE_TO_PYTHON: typing.Dict[str, typing.Dict[str, str]] = {
    BaseClassEnum.AVRO_MODEL.value: _AVRO_TYPE_TO_PYTHON,
    BaseClassEnum.PYDANTIC_MODEL.value: {
        **_AVRO_TYPE_TO_PYTHON,
        field_utils.UUID: "pydantic.UUID4",
    },
    BaseClassEnum.AVRO_DANTIC_MODEL.value: {
        **_AVRO_TYPE_TO_PYTHON,
        field_utils.UUID: "pydantic.UUID4",
    },
}

_LOGICAL_TYPES_IMPORTS: typing.Dict[str, str] = {
    field_utils.DECIMAL: "import decimal",
    field_utils.DATE: "import datetime",
    field_utils.TIME_MILLIS: "import datetime",
    field_utils.TIME_MICROS: "from dataclasses_avroschema import types",
    field_utils.TIMESTAMP_MILLIS: "import datetime",
    field_utils.TIMESTAMP_MICROS: "from dataclasses_avroschema import types",
    field_utils.UUID: "import uuid",
}
LOGICAL_TYPES_IMPORTS: typing.Dict[str, typing.Dict[str, str]] = {
    BaseClassEnum.AVRO_MODEL.value: _LOGICAL_TYPES_IMPORTS,
    BaseClassEnum.PYDANTIC_MODEL.value: {
        **_LOGICAL_TYPES_IMPORTS,
        field_utils.UUID: "import pydantic",
    },
    BaseClassEnum.AVRO_DANTIC_MODEL.value: {
        **_LOGICAL_TYPES_IMPORTS,
        field_utils.UUID: "import pydantic",
    },
}

# Avro types to python types
LOGICAL_TYPES_TO_PYTHON = {
    field_utils.DATE: lambda value: datetime.date.fromtimestamp(60 * 60 * 24 * value),
    field_utils.TIME_MILLIS: lambda value: (datetime.datetime.min + datetime.timedelta(milliseconds=value)).time(),
    field_utils.TIME_MICROS: lambda value: (datetime.datetime.min + datetime.timedelta(microseconds=value)).time(),
    field_utils.TIMESTAMP_MILLIS: lambda value: datetime.datetime.fromtimestamp(value / 1000),
    field_utils.TIMESTAMP_MICROS: lambda value: datetime.datetime.fromtimestamp(value / 1000000),
}

# Logical types objects to template
LOGICAL_TYPE_TEMPLATES = {
    field_utils.DATE: lambda date_obj: templates.date_template.safe_substitute(
        year=date_obj.year, month=date_obj.month, day=date_obj.day
    ),
    field_utils.TIME_MILLIS: lambda time_obj: templates.time_template.safe_substitute(
        hour=time_obj.hour, minute=time_obj.minute, second=time_obj.second
    ),
    field_utils.TIME_MICROS: lambda time_obj: templates.time_micros_template.safe_substitute(
        hour=time_obj.hour, minute=time_obj.minute, second=time_obj.second, microsecond=time_obj.microsecond
    ),
    field_utils.TIMESTAMP_MILLIS: lambda datetime_obj: templates.datetime_template.safe_substitute(
        year=datetime_obj.year,
        month=datetime_obj.month,
        day=datetime_obj.day,
        hour=datetime_obj.hour,
        minute=datetime_obj.minute,
        second=datetime_obj.second,
    ),
    field_utils.TIMESTAMP_MICROS: lambda datetime_obj: templates.datetime_micros_template.safe_substitute(
        year=datetime_obj.year,
        month=datetime_obj.month,
        day=datetime_obj.day,
        hour=datetime_obj.hour,
        minute=datetime_obj.minute,
        second=datetime_obj.second,
        microsecond=datetime_obj.microsecond,
    ),
}


def render_datetime(*, value: int, format: str) -> str:
    fn = LOGICAL_TYPES_TO_PYTHON[format]
    datetime_obj = fn(value)

    template = LOGICAL_TYPE_TEMPLATES[format]
    return template(datetime_obj)



from typing import Any, Dict, Type, TypeVar

from fastavro.validation import validate

from .schema_generator import AvroModel, JsonDict

try:
    from pydantic import BaseModel  # pragma: no cover
except ImportError as ex:  # pragma: no cover
    raise Exception("pydantic must be installed in order to use AvroBaseModel") from ex  # pragma: no cover

CT = TypeVar("CT", bound="AvroBaseModel")


class AvroBaseModel(BaseModel, AvroModel):  # type: ignore
    @classmethod
    def json_schema(cls: Type[CT], *args: Any, **kwargs: Any) -> str:
        return cls.schema_json(*args, **kwargs)

    def asdict(self) -> JsonDict:
        """
        Document this. asdict vs dict
        """
        data = self.dict()

        # te standardize called can be replaced if we have a custom implementation of asdict
        # for now I think is better to use the native implementation
        return {key: self.standardize_custom_type(value) for key, value in data.items()}

    def validate_avro(self) -> bool:
        """
        Document this!!!
        """
        schema = self.avro_schema_to_python()
        return validate(self.asdict(), schema)

    @classmethod
    def parse_obj(cls: Type["AvroBaseModel"], data: Dict) -> "AvroBaseModel":
        return super().parse_obj(data)

    @classmethod
    def fake(cls: Type[CT], **data: Dict[str, Any]) -> "AvroBaseModel":
        """
        Creates a fake instance of the model.

        Attributes:
            data: Dict[str, Any] represent the user values to use in the instance
        """
        # only generate fakes for fields that were not provided in data
        payload = {field.name: field.fake() for field in cls.get_fields() if field.name not in data.keys()}
        payload.update(data)

        return cls.parse_obj(payload)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" setup.py for dataclasses-avroschema."""

from setuptools import find_packages, setup

__version__ = "0.37.4"


with open("README.md") as readme_file:
    long_description = readme_file.read()

setup(
    name="dataclasses-avroschema",
    version=__version__,
    description="Generate Avro Schemas from a Python class",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Marcos Schroh",
    install_requires=[
        "inflect>=5.3.0",
        "fastavro>=1.7.0",
        "pytz",
        "dacite>=1.8.0",
        "faker>=17.6.0",
        'typing_extensions>=3.7.4;python_version<"3.9"',
        "stringcase>=1.2.0",
    ],
    extras_require={
        "pydantic": ["pydantic>=1.9.0"],
        "cli": ["dc-avro>=0.6.1"],
    },
    author_email="schrohm@gmail.com",
    url="https://github.com/marcosschroh/dataclasses-avroschema",
    download_url="",
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development",
    ],
    keywords=(
        """
        Python, Data Classes, Avro Schema, Avro, Apache, Data Streaming
        """
    ),
)



import dataclasses
import typing

import pytest

from dataclasses_avroschema import field_utils, fields

from . import consts


@pytest.mark.parametrize("primitive_type", field_utils.PYTHON_INMUTABLE_TYPES)
def test_primitive_types(primitive_type):
    name = "a_field"
    field = fields.AvroField(name, primitive_type, default=dataclasses.MISSING)
    avro_type = field_utils.PYTHON_TYPE_TO_AVRO[primitive_type]

    assert {"name": name, "type": avro_type} == field.to_dict()


@pytest.mark.parametrize("primitive_type", field_utils.PYTHON_INMUTABLE_TYPES)
def test_primitive_types_with_default_value_none(primitive_type):
    name = "a_field"
    field = fields.AvroField(name, primitive_type, default=None)
    avro_type = [field_utils.NULL, field_utils.PYTHON_TYPE_TO_AVRO[primitive_type]]

    assert {"name": name, "type": avro_type, "default": None} == field.to_dict()


@pytest.mark.parametrize("primitive_type,default", consts.PRIMITIVE_TYPES_AND_DEFAULTS)
def test_primitive_types_with_default_value(primitive_type, default):
    name = "a_field"
    field = fields.AvroField(name, primitive_type, default=default)
    avro_type = field_utils.PYTHON_TYPE_TO_AVRO[primitive_type]

    if primitive_type is bytes:
        default = default.decode()

    assert {"name": name, "type": avro_type, "default": default} == field.to_dict()


@pytest.mark.parametrize("primitive_type,invalid_default", consts.PRIMITIVE_TYPES_AND_INVALID_DEFAULTS)
def test_invalid_default_values(primitive_type, invalid_default):
    name = "a_field"
    field = fields.AvroField(name, primitive_type, default=invalid_default)

    msg = f"Invalid default type. Default should be {primitive_type}"
    with pytest.raises(AssertionError, match=msg):
        field.to_dict()


def test_invalid_type():
    msg = f"Type {typing.Any} is unknown. Please check the valid types at https://marcosschroh.github.io/dataclasses-avroschema/fields_specification/#avro-field-and-python-types-summary"  # noqa
    with pytest.raises(ValueError, match=msg):
        name = "a_field"
        fields.AvroField(name, typing.Any)



import abc
import dataclasses
import typing
from collections import OrderedDict

from dataclasses_avroschema import utils
from dataclasses_avroschema.fields import AvroField, FieldType


@dataclasses.dataclass  # type: ignore
class BaseSchemaDefinition(abc.ABC):
    """
    Minimal Schema definition
    """

    __slots__ = (
        "type",
        "klass",
        "parent",
        "matadata",
    )

    type: str
    klass: typing.Any
    parent: typing.Any
    metadata: utils.SchemaMetadata

    @abc.abstractmethod
    def get_rendered_fields(self) -> typing.List[OrderedDict]:
        ...  # pragma: no cover

    @abc.abstractmethod
    def render(self) -> OrderedDict:
        ...  # pragma: no cover

    def get_schema_name(self) -> str:
        return self.klass.metadata.schema_name or self.klass.__name__

    def generate_documentation(self) -> typing.Optional[str]:
        if isinstance(self.metadata.schema_doc, str):
            doc = self.metadata.schema_doc
        else:
            doc = self.klass.__doc__

        if doc is not None:
            return doc.replace("\n", "")
        return None


@dataclasses.dataclass
class AvroSchemaDefinition(BaseSchemaDefinition):
    fields: typing.List[FieldType] = dataclasses.field(default_factory=list)

    def __post_init__(self) -> None:
        self.fields = self.parse_dataclasses_fields()

    def parse_dataclasses_fields(self) -> typing.List[FieldType]:
        if utils.is_faust_model(self.klass):
            return self.parse_faust_fields()
        elif utils.is_pydantic_model(self.klass):
            return self.parse_pydantic_fields()
        return self.parse_fields()

    def parse_fields(self) -> typing.List[FieldType]:
        return [
            AvroField(
                dataclass_field.name,
                dataclass_field.type,
                default=dataclass_field.default,
                default_factory=dataclass_field.default_factory,  # type: ignore  # TODO: resolve mypy
                metadata=dataclass_field.metadata,
                model_metadata=self.metadata,
                parent=self.parent,
            )
            for dataclass_field in dataclasses.fields(self.klass)
        ]

    def parse_faust_fields(self) -> typing.List[FieldType]:
        schema_fields = []

        for dataclass_field in dataclasses.fields(self.klass):
            faust_field = dataclass_field.default
            metadata = dataclass_field.metadata
            default_factory = dataclasses.MISSING

            if faust_field is not dataclasses.MISSING:
                if faust_field.required:
                    default = dataclasses.MISSING
                else:
                    default = faust_field.default

                    if isinstance(default, dataclasses.Field):
                        metadata = default.metadata
                        default_factory = default.default_factory  # type: ignore  # TODO: resolve mypy
                        default = dataclasses.MISSING

                schema_fields.append(
                    AvroField(
                        dataclass_field.name,
                        dataclass_field.type,
                        default=default,
                        default_factory=default_factory,
                        metadata=metadata,
                        model_metadata=self.metadata,
                        parent=self.parent,
                    )
                )

        return schema_fields

    def parse_pydantic_fields(self) -> typing.List[FieldType]:
        return [
            AvroField(
                model_field.name,
                model_field.annotation,
                default=dataclasses.MISSING
                if model_field.required or model_field.default_factory
                else model_field.default,
                default_factory=model_field.default_factory,
                metadata=getattr(model_field, "metadata", {}),
                model_metadata=self.metadata,
                parent=self.parent,
            )
            for model_field in self.klass.__fields__.values()
        ]

    def get_rendered_fields(self) -> typing.List[OrderedDict]:
        return [field.render() for field in self.fields]

    def render(self) -> OrderedDict:
        schema = OrderedDict(
            [
                ("type", self.type),
                ("name", self.get_schema_name()),
                ("fields", self.get_rendered_fields()),
            ]
        )

        if self.metadata.schema_doc:
            doc = self.generate_documentation()
            if doc is not None:
                schema["doc"] = doc

        if self.metadata.namespace is not None:
            schema["namespace"] = self.metadata.namespace

        if self.metadata.aliases is not None:
            schema["aliases"] = self.metadata.aliases

        return schema
